#!/bin/false

# Copyright (c) 2022 Vít Labuda. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#  3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from typing import Union, Type
from .RequestMessage import RequestMessage
from .SuccessfulResponseMessage import SuccessfulResponseMessage
from .ErroneousResponseMessage import ErroneousResponseMessage
from ._InternalV1Constants import _InternalV1Constants
from ._ValidationHelpers import _ValidationHelpers
from ..etc.UninstantiableClassMixin import UninstantiableClassMixin


__all__ = "WireformatParsingHelpers",


# This is not really a class as per OOP definition, but rather a collection of semi-independent functions.
class WireformatParsingHelpers(UninstantiableClassMixin):
    @classmethod
    def instantiate_appropriate_message_class_from_wireformat(cls, wireformat_bytes: bytes) -> Union[RequestMessage, SuccessfulResponseMessage, ErroneousResponseMessage]:  # DP: Factory
        """
        :raises InvalidMessageDataExc
        """

        return cls.detect_message_class_from_wireformat(wireformat_bytes).from_wireformat(wireformat_bytes)

    # Keep in mind that this method performs absolutely minimal validation (it is done this way so parts of code are
    #  not unnecessarily repeated) - the fact that this function successfully returns a message class does not mean
    #  that calling 'from_wireformat()' on it will not fail!
    @staticmethod
    def detect_message_class_from_wireformat(wireformat_bytes: bytes) -> Union[Type[RequestMessage], Type[SuccessfulResponseMessage], Type[ErroneousResponseMessage]]:
        """
        :raises InvalidMessageDataExc
        """

        _ValidationHelpers.validate_wireformat_size(wireformat_bytes)

        message_type_byte = int(wireformat_bytes[2])

        if message_type_byte & _InternalV1Constants.BITMASK_RESPONSE_BIT:  # = If 'response_bit' is set
            if message_type_byte & _InternalV1Constants.BITMASK_ERROR_BIT:  # = If both 'response_bit' and 'error_bit' are set
                return ErroneousResponseMessage
            return SuccessfulResponseMessage

        return RequestMessage



#!/bin/false

# Copyright (c) 2022 Vít Labuda. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#  3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from __future__ import annotations
from typing import Union
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address
from .WireformattableIface import WireformattableIface
from .MessageType import MessageType
from ._ValidationHelpers import _ValidationHelpers
from ._WireformatHelpers import _WireformatHelpers
from ..exc.InvalidWireformatMessageDataExc import InvalidWireformatMessageDataExc


__all__ = "SuccessfulResponseMessage",


@dataclass(frozen=True)
class SuccessfulResponseMessage(WireformattableIface):
    message_type: MessageType
    cache_lifetime: int
    message_identifier: int
    source_ip_address: Union[IPv4Address, IPv6Address]
    destination_ip_address: Union[IPv4Address, IPv6Address]

    def __post_init__(self):
        """
        :raises InvalidMessageDataExc
        """

        _ValidationHelpers.validate_cache_lifetime(cache_lifetime=self.cache_lifetime)

        _ValidationHelpers.validate_message_identifier(message_identifier=self.message_identifier)

        _ValidationHelpers.validate_ip_address_versions_for_message_type(
            message_type=self.message_type,
            source_ip_address=self.source_ip_address,
            destination_ip_address=self.destination_ip_address,
            expected_4to6_ip_version=IPv6Address,
            expected_6to4_ip_version=IPv4Address
        )

    @classmethod
    def from_wireformat(cls, wireformat_bytes: bytes) -> SuccessfulResponseMessage:  # DP: Factory
        """
        :raises InvalidMessageDataExc
        """

        parsed_wireformat = _WireformatHelpers.from_wireformat(
            wireformat_bytes=wireformat_bytes,
            expected_4to6_ip_version=IPv6Address,
            expected_6to4_ip_version=IPv4Address
        )

        assert ((parsed_wireformat.source_ip_address is not None) and (parsed_wireformat.destination_ip_address is not None))

        if (not parsed_wireformat.response_bit) or parsed_wireformat.error_bit or parsed_wireformat.icmp_bit:
            raise InvalidWireformatMessageDataExc("In case of a successful response message, 'response_bit' must be set, whereas 'error_bit' and 'icmp_bit' must be unset!")

        return cls(
            message_type=parsed_wireformat.message_type,
            cache_lifetime=parsed_wireformat.cache_lifetime,
            message_identifier=parsed_wireformat.message_identifier,
            source_ip_address=parsed_wireformat.source_ip_address,
            destination_ip_address=parsed_wireformat.destination_ip_address
        )

    def to_wireformat(self) -> bytes:
        return _WireformatHelpers.to_wireformat(
            response_bit=True,
            error_bit=False,
            icmp_bit=False,
            message_type=self.message_type,
            cache_lifetime=self.cache_lifetime,
            message_identifier=self.message_identifier,
            source_ip_address=self.source_ip_address,
            destination_ip_address=self.destination_ip_address
        )



#!/bin/false

# Copyright (c) 2022 Vít Labuda. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#  3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from __future__ import annotations
from typing import Union
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address
from .WireformattableIface import WireformattableIface
from .MessageType import MessageType
from .SuccessfulResponseMessage import SuccessfulResponseMessage
from .ErroneousResponseMessage import ErroneousResponseMessage
from ._ValidationHelpers import _ValidationHelpers
from ._WireformatHelpers import _WireformatHelpers
from ..exc.InvalidWireformatMessageDataExc import InvalidWireformatMessageDataExc


__all__ = "RequestMessage",


@dataclass(frozen=True)
class RequestMessage(WireformattableIface):
    message_type: MessageType
    message_identifier: int
    source_ip_address: Union[IPv4Address, IPv6Address]
    destination_ip_address: Union[IPv4Address, IPv6Address]

    def __post_init__(self):
        """
        :raises InvalidMessageDataExc
        """

        _ValidationHelpers.validate_message_identifier(message_identifier=self.message_identifier)

        _ValidationHelpers.validate_ip_address_versions_for_message_type(
            message_type=self.message_type,
            source_ip_address=self.source_ip_address,
            destination_ip_address=self.destination_ip_address,
            expected_4to6_ip_version=IPv4Address,
            expected_6to4_ip_version=IPv6Address
        )

    @classmethod
    def from_wireformat(cls, wireformat_bytes: bytes) -> RequestMessage:  # DP: Factory
        """
        :raises InvalidMessageDataExc
        """

        parsed_wireformat = _WireformatHelpers.from_wireformat(
            wireformat_bytes=wireformat_bytes,
            expected_4to6_ip_version=IPv4Address,
            expected_6to4_ip_version=IPv6Address
        )

        assert ((parsed_wireformat.source_ip_address is not None) and (parsed_wireformat.destination_ip_address is not None))

        if parsed_wireformat.response_bit or parsed_wireformat.error_bit or parsed_wireformat.icmp_bit:
            raise InvalidWireformatMessageDataExc("In case of a request message, 'response_bit', 'error_bit' and 'icmp_bit' must all be unset!")

        return cls(
            message_type=parsed_wireformat.message_type,
            message_identifier=parsed_wireformat.message_identifier,
            source_ip_address=parsed_wireformat.source_ip_address,
            destination_ip_address=parsed_wireformat.destination_ip_address
        )

    def to_wireformat(self) -> bytes:
        return _WireformatHelpers.to_wireformat(
            response_bit=False,
            error_bit=False,
            icmp_bit=False,
            message_type=self.message_type,
            cache_lifetime=0,
            message_identifier=self.message_identifier,
            source_ip_address=self.source_ip_address,
            destination_ip_address=self.destination_ip_address
        )

    def generate_successful_response(self, cache_lifetime: int, source_ip_address: Union[IPv4Address, IPv6Address], destination_ip_address: Union[IPv4Address, IPv6Address]) -> SuccessfulResponseMessage:  # DP: Factory
        """
        :raises InvalidMessageDataExc
        """

        return SuccessfulResponseMessage(
            message_type=self.message_type,
            cache_lifetime=cache_lifetime,
            message_identifier=self.message_identifier,
            source_ip_address=source_ip_address,
            destination_ip_address=destination_ip_address
        )

    def generate_erroneous_response(self, icmp_bit: bool) -> ErroneousResponseMessage:  # DP: Factory
        """
        :raises InvalidMessageDataExc
        """

        return ErroneousResponseMessage(
            icmp_bit=icmp_bit,
            message_type=self.message_type,
            message_identifier=self.message_identifier
        )



#!/bin/false

# Copyright (c) 2022 Vít Labuda. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#  3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from typing import final, Final
from ..etc.UninstantiableClassMixin import UninstantiableClassMixin


@final
class _InternalV1Constants(UninstantiableClassMixin):
    BITMASK_RESPONSE_BIT: Final[int] = 0x80
    BITMASK_ERROR_BIT: Final[int] = 0x40
    BITMASK_ICMP_BIT: Final[int] = 0x20
    BITMASK_MESSAGE_TYPE_INT: Final[int] = 0x1f



#!/bin/false

# Copyright (c) 2022 Vít Labuda. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#  3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from typing import Optional, Union, Type
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address
import struct
from .MessageType import MessageType
from .V1Constants import V1Constants
from ._InternalV1Constants import _InternalV1Constants
from ._MiscHelpers import _MiscHelpers
from ._ValidationHelpers import _ValidationHelpers
from ..TundraXaxlibConstants import TundraXaxlibConstants
from ..etc.UninstantiableClassMixin import UninstantiableClassMixin
from ..exc.InvalidWireformatMessageDataExc import InvalidWireformatMessageDataExc


# This is not really a class as per OOP definition, but rather a collection of semi-independent functions.
class _WireformatHelpers(UninstantiableClassMixin):
    @dataclass(frozen=True)
    class ParsedWireformat:
        response_bit: bool
        error_bit: bool
        icmp_bit: bool
        message_type: MessageType
        cache_lifetime: int
        message_identifier: int
        source_ip_address: Optional[Union[IPv4Address, IPv6Address]]
        destination_ip_address: Optional[Union[IPv4Address, IPv6Address]]

    @classmethod
    def from_wireformat(cls, wireformat_bytes: bytes, expected_4to6_ip_version: Optional[Union[Type[IPv4Address], Type[IPv6Address]]], expected_6to4_ip_version: Optional[Union[Type[IPv4Address], Type[IPv6Address]]]) -> ParsedWireformat:
        """
        :raises InvalidMessageDataExc
        """

        _ValidationHelpers.validate_wireformat_size(wireformat_bytes)

        first_8_bytes = wireformat_bytes[0:8]
        source_ip_address_bytes = wireformat_bytes[8:24]
        destination_ip_address_bytes = wireformat_bytes[24:40]

        magic_byte, protocol_version, message_type_byte, cache_lifetime, message_identifier = struct.unpack("!cBBBI", first_8_bytes)

        if magic_byte != TundraXaxlibConstants.MAGIC_BYTE:
            raise InvalidWireformatMessageDataExc(f"'magic_byte' must be {TundraXaxlibConstants.MAGIC_BYTE}, got {magic_byte}!")
        if protocol_version != V1Constants.PROTOCOL_VERSION:
            raise InvalidWireformatMessageDataExc(f"'protocol_version' must be {V1Constants.PROTOCOL_VERSION}, got {protocol_version}!")

        response_bit, error_bit, icmp_bit, message_type = cls._convert_message_type_byte_to_bits_and_message_type(message_type_byte)

        ip_version = _MiscHelpers.get_expected_ip_version_for_message_type(
            message_type=message_type,
            expected_4to6_ip_version=expected_4to6_ip_version,
            expected_6to4_ip_version=expected_6to4_ip_version
        )

        source_ip_address = cls._convert_16_bytes_to_optional_ip_address_object(source_ip_address_bytes, ip_version)
        destination_ip_address = cls._convert_16_bytes_to_optional_ip_address_object(destination_ip_address_bytes, ip_version)

        return cls.ParsedWireformat(
            response_bit=response_bit,
            error_bit=error_bit,
            icmp_bit=icmp_bit,
            message_type=message_type,
            cache_lifetime=cache_lifetime,
            message_identifier=message_identifier,
            source_ip_address=source_ip_address,
            destination_ip_address=destination_ip_address
        )

    @staticmethod
    def _convert_message_type_byte_to_bits_and_message_type(message_type_byte: int) -> tuple[bool, bool, bool, MessageType]:
        """
        :raises InvalidMessageDataExc
        """

        response_bit = bool(message_type_byte & _InternalV1Constants.BITMASK_RESPONSE_BIT)
        error_bit = bool(message_type_byte & _InternalV1Constants.BITMASK_ERROR_BIT)
        icmp_bit = bool(message_type_byte & _InternalV1Constants.BITMASK_ICMP_BIT)

        message_type_int = (message_type_byte & _InternalV1Constants.BITMASK_MESSAGE_TYPE_INT)
        try:
            message_type = ({member.value: member for member in MessageType})[message_type_int]
        except KeyError:
            raise InvalidWireformatMessageDataExc(f"'message_type' must be of one of these values: {repr(tuple(member.value for member in MessageType))}, got {message_type_int}!")

        return response_bit, error_bit, icmp_bit, message_type

    @staticmethod
    def _convert_16_bytes_to_optional_ip_address_object(ip_address_bytes: bytes, ip_version: Optional[Union[Type[IPv4Address], Type[IPv6Address]]]) -> Optional[Union[IPv4Address, IPv6Address]]:
        """
        :raises InvalidMessageDataExc
        """

        assert (len(ip_address_bytes) == 16)

        if isinstance(ip_version, type):  # issubclass() does not accept 'None' (or any other instance which is not a class) as its first argument
            if issubclass(ip_version, IPv4Address):
                if ip_address_bytes[4:] != (b'\x00' * 12):
                    raise InvalidWireformatMessageDataExc("'source_ip_address' or 'destination_ip_address' should be carrying an IPv4 address, but the field's last 12 bytes are nonzero!")
                return IPv4Address(ip_address_bytes[0:4])

            if issubclass(ip_version, IPv6Address):
                return IPv6Address(ip_address_bytes)

        return None

    @classmethod
    def to_wireformat(cls, response_bit: bool, error_bit: bool, icmp_bit: bool,
                      message_type: MessageType, cache_lifetime: int, message_identifier: int,
                      source_ip_address: Optional[Union[IPv4Address, IPv6Address]],
                      destination_ip_address: Optional[Union[IPv4Address, IPv6Address]]) -> bytes:

        message_type_byte = cls._convert_bits_and_message_type_to_message_type_byte(response_bit, error_bit, icmp_bit, message_type)
        first_8_bytes = struct.pack("!cBBBI",
                                    TundraXaxlibConstants.MAGIC_BYTE,
                                    V1Constants.PROTOCOL_VERSION,
                                    message_type_byte,
                                    cache_lifetime,
                                    message_identifier
                                    )

        source_ip_address_bytes = cls._convert_optional_ip_address_object_to_16_bytes(source_ip_address)
        destination_ip_address_bytes = cls._convert_optional_ip_address_object_to_16_bytes(destination_ip_address)

        result = (first_8_bytes + source_ip_address_bytes + destination_ip_address_bytes)
        assert (len(result) == V1Constants.WIREFORMAT_MESSAGE_SIZE)
        return result

    @staticmethod
    def _convert_bits_and_message_type_to_message_type_byte(response_bit: bool, error_bit: bool, icmp_bit: bool, message_type: MessageType) -> int:
        message_type_byte = message_type.value
        if response_bit:
            message_type_byte |= _InternalV1Constants.BITMASK_RESPONSE_BIT
        if error_bit:
            message_type_byte |= _InternalV1Constants.BITMASK_ERROR_BIT
        if icmp_bit:
            message_type_byte |= _InternalV1Constants.BITMASK_ICMP_BIT

        return message_type_byte

    @staticmethod
    def _convert_optional_ip_address_object_to_16_bytes(ip_address_object: Optional[Union[IPv4Address, IPv6Address]]) -> bytes:
        if ip_address_object is None:
            return b'\x00' * 16

        return (ip_address_object.packed + (b'\x00' * 12))[0:16]



#!/bin/false

# Copyright (c) 2022 Vít Labuda. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#  3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import sys
import os
_REPOSITORY_ROOT_DIR = os.path.dirname(os.path.realpath(__file__))
if _REPOSITORY_ROOT_DIR not in sys.path:
    sys.path.insert(0, _REPOSITORY_ROOT_DIR)
os.chdir(_REPOSITORY_ROOT_DIR)

import setuptools
from tundra_xaxlib.TundraXaxlibConstants import TundraXaxlibConstants


with open("./README.md", "r") as readme_io:
    readme_text = readme_io.read()


setuptools.setup(
    name="tundra-xaxlib-python",
    version=TundraXaxlibConstants.LIBRARY_VERSION,
    description="Tundra-NAT64 external address translation library for Python",
    long_description=readme_text,
    long_description_content_type="text/markdown",
    author="Vít Labuda",
    author_email="dev@vitlabuda.cz",
    license="BSD License",
    url="https://github.com/vitlabuda/tundra-xaxlib-python",
    project_urls={
        "Bug Tracker": "https://github.com/vitlabuda/tundra-xaxlib-python/issues",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries",
        "Typing :: Typed",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.9"
)



#!/bin/false

# Copyright (c) 2022 Vít Labuda. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#  1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#  3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from typing import Union, Type
from ipaddress import IPv4Address, IPv6Address
from .V1Constants import V1Constants
from .MessageType import MessageType
from ._MiscHelpers import _MiscHelpers
from ..etc.UninstantiableClassMixin import UninstantiableClassMixin
from ..exc.InvalidMessageDataExc import InvalidMessageDataExc
from ..exc.InvalidWireformatMessageDataExc import InvalidWireformatMessageDataExc


# This is not really a class as per OOP definition, but rather a collection of independent functions.
class _ValidationHelpers(UninstantiableClassMixin):
    @staticmethod
    def validate_wireformat_size(wireformat_bytes: bytes) -> None:
        """
        :raises InvalidMessageDataExc
        """

        wireformat_msg_len = len(wireformat_bytes)
        if wireformat_msg_len != V1Constants.WIREFORMAT_MESSAGE_SIZE:
            raise InvalidWireformatMessageDataExc(f"Wireformat messages must be exactly {V1Constants.WIREFORMAT_MESSAGE_SIZE} bytes in size, got {wireformat_msg_len} bytes!")

    @staticmethod
    def validate_icmp_bit_for_message_type(message_type: MessageType, icmp_bit: bool) -> None:
        """
        :raises InvalidMessageDataExc
        """

        if icmp_bit and (message_type not in (MessageType.MT_4TO6_MAIN_PACKET, MessageType.MT_6TO4_MAIN_PACKET)):
            raise InvalidMessageDataExc(f"'icmp_bit' may be set only for the following message types: {MessageType.MT_4TO6_MAIN_PACKET.name}, {MessageType.MT_6TO4_MAIN_PACKET.name} (got {message_type.name})")

    @staticmethod
    def validate_cache_lifetime(cache_lifetime: int) -> None:
        """
        :raises InvalidMessageDataExc
        """

        if (cache_lifetime < 0) or (cache_lifetime > 255):
            raise InvalidMessageDataExc(f"'cache_lifetime' is out of range (0 - 255): {cache_lifetime}")

    @staticmethod
    def validate_message_identifier(message_identifier: int) -> None:
        """
        :raises InvalidMessageDataExc
        """

        if (message_identifier < 0) or (message_identifier > 4_294_967_295):  # 'message_identifier' is uint32_t
            raise InvalidMessageDataExc(f"'message_identifier' is out of range (0 - 4_294_967_295): {message_identifier}")

    @staticmethod
    def validate_ip_address_versions_for_message_type(message_type: MessageType,
                                                      source_ip_address: Union[IPv4Address, IPv6Address],
                                                      destination_ip_address: Union[IPv4Address, IPv6Address],
                                                      expected_4to6_ip_version: Union[Type[IPv4Address], Type[IPv6Address]],
                                                      expected_6to4_ip_version: Union[Type[IPv4Address], Type[IPv6Address]]) -> None:
        """
        :raises InvalidMessageDataExc
        """

        ip_version = _MiscHelpers.get_expected_ip_version_for_message_type(
            message_type=message_type,
            expected_4to6_ip_version=expected_4to6_ip_version,
            expected_6to4_ip_version=expected_6to4_ip_version
        )
        assert (ip_version is not None)

        if (not isinstance(source_ip_address, ip_version)) or (not isinstance(destination_ip_address, ip_version)):
            raise InvalidMessageDataExc(f"If 'message_type' is {message_type.name}, the source & destination IP must be an {ip_version.__name__}!")
